﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

using CMS.DataEngine;
using CMS.DataEngine.Query;
using CMS.DocumentEngine;
using CMS.Ecommerce;
using CMS.FormEngine;
using CMS.Helpers;
using CMS.SiteProvider;

namespace Kentico
{
    public sealed class AbsolutePageLinksDetectionUtility
    {
        private const int PAGE_SIZE = 500;

        private Regex AbsoluteUrlRegex;
        private readonly LogService logService;
        private DetectionState state = new DetectionState();
        private bool absoluteLinksDetected = false;

        public AbsolutePageLinksDetectionUtility(LogService logService)
        {
            this.logService = logService ?? throw new ArgumentNullException(nameof(logService));
        }


        public void DetectPotentialIssues()
        {
            logService.LogMessage("Detection started.");
            PrepareSKUFields();
            var sites = SiteInfoProvider.GetSites().WhereTrue("SiteIsContentOnly").WhereNotEmpty("SitePresentationURL");

            foreach (var site in sites)
            {
                state.Site = site;
                logService.LogMessage($"\n\nSite {site.DisplayName}");

                AbsoluteUrlRegex = PrepareLinkRegex();

                ExaminePages();
                ExamineSKUs();
            }

            ExamineGlobalSKUs();

            if (!absoluteLinksDetected)
            {
                logService.LogMessage("\nNo pages/products with absolute page links found.");
            }

            logService.LogMessage("\nDetection finished.");
        }


        private Regex PrepareLinkRegex()
        {
            string pattern = Regex.Escape(state.Site.SitePresentationURL.TrimEnd('/'));
            pattern = $"href=[\"']?({pattern}/?)";
            return RegexHelper.GetRegex(pattern, RegexHelper.DefaultOptions | RegexOptions.IgnoreCase);
        }


        private void PrepareSKUFields()
        {
            var skuDataClass = DataClassInfoProvider.GetDataClassInfo(PredefinedObjectType.SKU);
            var form = new FormInfo(skuDataClass.ClassFormDefinition);
            state.SkuFields = form.GetFields(FormFieldControlTypeEnum.HtmlAreaControl).Select(f => f.Name);
        }


        private void ExamineGlobalSKUs()
        {
            logService.LogMessage("\nSuspected standalone global SKUs:");

            // Standalone global SKUs
            var standaloneGlobalSKUs = new ObjectQuery<SKUInfo>()
                .OnlyGlobal()
                .WhereNotIn("SKUID", SKUInfoProvider.GetSKUs()
                    .Column("NodeSKUID")
                    .From(SystemViewNames.View_CMS_Tree_Joined)
                    .WhereNotNull("NodeSKUID"));

            standaloneGlobalSKUs.ForEachPage(q =>
            {
                foreach (var sku in q)
                {
                    ExamineFields(sku, state.SkuFields);

                }
            }, PAGE_SIZE);
        }


        private void ExaminePages()
        {
            string siteName = state.Site.SiteName;

            logService.LogMessage("\nSuspected pages:");

            // Get page types containing Rich text editor (using simple heuristics)
            var classes = DataClassInfoProvider
                .GetClasses()
                .WhereTrue("ClassIsContentOnly")
                .WhereTrue("ClassIsDocumentType")
                .WhereContains("ClassFormDefinition", "HtmlAreaControl")
                .ToList();

            foreach (var pageType in classes)
            {
                var form = new FormInfo(pageType.ClassFormDefinition);
                var fields = form.GetFields(FormFieldControlTypeEnum.HtmlAreaControl).Select(f => f.Name);

                var cultureVersions = CultureSiteInfoProvider.GetSiteCultureCodes(siteName);

                foreach (string culture in cultureVersions)
                {
                    state.Culture = culture;

                    var latestVersionsQuery = new DocumentQuery(pageType.ClassName)
                        .OnSite(siteName)
                        .Culture(culture)
                        .CombineWithDefaultCulture(combine: false)
                        .LatestVersion()
                        // Get pages that are not in the published step
                        .WhereNotEquals("DocumentCheckedOutVersionHistoryID", "COALESCE(DocumentPublishedVersionHistoryID, 0)".AsExpression())
                        .OrderBy("NodeAliasPath");

                    latestVersionsQuery.ForEachPage(q =>
                    {
                        foreach (var page in q)
                        {
                            ExamineFields(page, fields);
                        }
                    }, PAGE_SIZE);

                    var publishedVersionQuery = new DocumentQuery(pageType.ClassName)
                        .OnSite(siteName)
                        .Culture(culture)
                        .CombineWithDefaultCulture(combine: false)
                        .PublishedVersion()
                        .OrderBy("NodeAliasPath"); ;

                    publishedVersionQuery.ForEachPage(q =>
                    {
                        foreach (var page in q)
                        {
                            ExamineFields(page, fields);
                        }
                    }, PAGE_SIZE);
                }
            }
        }


        private void ExamineSKUs()
        {
            logService.LogMessage("\nSuspected SKUs:");

            string siteName = state.Site.SiteName;

            var cultureVersions = CultureSiteInfoProvider.GetSiteCultureCodes(siteName);

            // Update latest version of pages as SKU data are contained in latest page's data.
            foreach (string culture in cultureVersions)
            {
                state.Culture = culture;

                var latestPagesWithSku = new DocumentQuery()
                    .WhereNotNull("NodeSKUID")
                    .OnSite(siteName)
                    .Culture(culture)
                    .CombineWithDefaultCulture(combine: false)
                    .LatestVersion()
                    // Get pages that are not in the published step
                    .WhereNotEquals("DocumentCheckedOutVersionHistoryID", "COALESCE(DocumentPublishedVersionHistoryID, 0)".AsExpression())
                    .OrderBy("NodeAliasPath");

                latestPagesWithSku.ForEachPage(q =>
                {
                    foreach (var pageSku in q)
                    {
                        pageSku.MakeComplete(loadFromDb: true);

                        if (pageSku is SKUTreeNode skuTreeNode)
                        {
                            ExamineFields(skuTreeNode, state.SkuFields);
                        }
                    }
                }, PAGE_SIZE);


                var publishedPagesWithSku = new DocumentQuery()
                    .WhereNotNull("NodeSKUID")
                    .OnSite(siteName)
                    .Culture(culture)
                    .CombineWithDefaultCulture(combine: false)
                    .PublishedVersion()
                    .OrderBy("NodeAliasPath");

                publishedPagesWithSku.ForEachPage(q =>
                {
                    foreach (var pageSku in q)
                    {
                        pageSku.MakeComplete(loadFromDb: true);

                        if (pageSku is SKUTreeNode skuTreeNode)
                        {
                            ExamineFields(skuTreeNode, state.SkuFields);
                        }
                    }
                }, PAGE_SIZE);
            }

            logService.LogMessage("\nSuspected standalone SKUs:");

            // Standalone SKUs
            var standaloneSKUs = new ObjectQuery<SKUInfo>()
                .OnSite(siteName)
                .WhereNotIn("SKUID", SKUInfoProvider.GetSKUs()
                    .Column("NodeSKUID")
                    .From(SystemViewNames.View_CMS_Tree_Joined)
                    .WhereNotNull("NodeSKUID").And().WhereEquals("NodeSiteID", state.Site.SiteID));

            standaloneSKUs.ForEachPage(q =>
            {
                foreach (var sku in q)
                {
                    ExamineFields(sku, state.SkuFields);
                }
            }, PAGE_SIZE);
        }


        private void FindAbsoluteLinks(BaseInfo info, string fieldName)
        {
            var currentValue = info.GetStringValue(fieldName, null);

            if (currentValue == null)
            {
                return;
            }

            if (AbsoluteUrlRegex.IsMatch(currentValue))
            {
                LogAbsoluteLink(info, fieldName);
                absoluteLinksDetected = true;
            }
        }


        private void ExamineFields(BaseInfo info, IEnumerable<string> fields)
        {
            foreach (var field in fields)
            {
                FindAbsoluteLinks(info, field);
            }
        }


        private void LogAbsoluteLink(BaseInfo info, string fieldName)
        {
            if (info is TreeNode treeNode)
            {
                logService.LogMessage($"Page '{treeNode.NodeAliasPath}' field '{fieldName}' in culture '{state.Culture}'");
            }

            if (info is SKUInfo sku)
            {
                logService.LogMessage($"SKU '{sku.SKUName}' field '{fieldName}'");
            }
        }


        private class DetectionState
        {
            public SiteInfo Site { get; set; }

            public string Culture { get; set; }

            public IEnumerable<string> SkuFields { get; set; }
        }
    }
}
